#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import dill as pickle
import shutil
import numpy as np
import pandas as pd

from {{package_name}} import utils
from {{package_name}}.monitoring.model_explainer import LimeExplainer
from {{package_name}}.models_training.models_sklearn.model_tfidf_svm import ModelTfidfSvm


# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)


class ModelExplainerTest(unittest.TestCase):
    '''Main class to test model_explainer'''

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)
        # Check if data folder exists
        data_path = utils.get_data_path()
        if not os.path.exists(data_path):
            os.mkdir(data_path)
        # Create a mock embedding
        fake_embedding = {'toto': [0.25, 0.90, 0.12], 'titi': [0.85, 0.12, 0.8], 'test': [0.5, 0.6, 0.1]}
        fake_path = os.path.join(data_path, 'fake_embedding.pkl')
        if os.path.exists(fake_path):
            os.remove(fake_path)
        with open(fake_path, 'wb') as f:
            pickle.dump(fake_embedding, f, pickle.HIGHEST_PROTOCOL)

    def test01_lime_explainer_nominal(self):
        '''Test of the Lime explainer'''

        # Model dir
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        # fake model_conf
        model_conf = {'preprocess_str': 'no_preprocess'}
        # Set vars
        x_train = np.array(["ceci est un test", "pas cela", "cela non plus", "ici test", "là, rien!"] * 100)
        y_train_mono_2 = np.array(['y_0', 'y_1', 'y_0', 'y_1', 'y_1'] * 100)
        y_train_mono_3 = np.array(['y_0', 'y_1', 'y_0', 'y_1', 'y_2'] * 100)
        y_train_multi = pd.DataFrame({'test1': [0, 0, 0, 1, 0] * 100, 'test2': [1, 0, 0, 0, 0] * 100, 'test3': [0, 0, 0, 1, 0] * 100})
        y_test_multi = y_train_multi.copy()

        # Mono-label - Mono class
        model = ModelTfidfSvm(model_dir=model_dir, multi_label=False)
        model.fit(x_train, y_train_mono_2)
        explainer = LimeExplainer(model, model_conf)
        explanation = explainer.explain_instance(content="ceci est un test", class_or_label_index=0)
        html = explainer.explain_instance_as_html(content="ceci est un test", class_or_label_index=0)
        exp_list = explainer.explain_instance_as_list(content="ceci est un test", class_or_label_index=0)
        explanation = explainer.explain_instance(content="ceci est un test", class_or_label_index=1)
        html = explainer.explain_instance_as_html(content="ceci est un test", class_or_label_index=1)
        exp_list = explainer.explain_instance_as_list(content="ceci est un test", class_or_label_index=1)
        explanation = explainer.explain_instance(content="ceci est un test", class_or_label_index=None)
        html = explainer.explain_instance_as_html(content="ceci est un test", class_or_label_index=None)
        exp_list = explainer.explain_instance_as_list(content="ceci est un test", class_or_label_index=None)
        remove_dir(model_dir)

        # Mono-label - Multi classes
        model = ModelTfidfSvm(model_dir=model_dir, multi_label=False)
        model.fit(x_train, y_train_mono_3)
        explainer = LimeExplainer(model, model_conf)
        explanation = explainer.explain_instance(content="ceci est un test", class_or_label_index=0)
        html = explainer.explain_instance_as_html(content="ceci est un test", class_or_label_index=0)
        exp_list = explainer.explain_instance_as_list(content="ceci est un test", class_or_label_index=0)
        explanation = explainer.explain_instance(content="ceci est un test", class_or_label_index=1)
        html = explainer.explain_instance_as_html(content="ceci est un test", class_or_label_index=1)
        exp_list = explainer.explain_instance_as_list(content="ceci est un test", class_or_label_index=1)
        explanation = explainer.explain_instance(content="ceci est un test", class_or_label_index=2)
        html = explainer.explain_instance_as_html(content="ceci est un test", class_or_label_index=2)
        exp_list = explainer.explain_instance_as_list(content="ceci est un test", class_or_label_index=2)
        explanation = explainer.explain_instance(content="ceci est un test", class_or_label_index=None)
        html = explainer.explain_instance_as_html(content="ceci est un test", class_or_label_index=None)
        exp_list = explainer.explain_instance_as_list(content="ceci est un test", class_or_label_index=None)
        remove_dir(model_dir)

        # Multi-labels
        model = ModelTfidfSvm(model_dir=model_dir, multi_label=True)
        model.fit(x_train, y_train_multi)
        explainer = LimeExplainer(model, model_conf)
        explanation = explainer.explain_instance(content="ceci est un test", class_or_label_index=0)
        html = explainer.explain_instance_as_html(content="ceci est un test", class_or_label_index=0)
        exp_list = explainer.explain_instance_as_list(content="ceci est un test", class_or_label_index=0)
        explanation = explainer.explain_instance(content="ceci est un test", class_or_label_index=1)
        html = explainer.explain_instance_as_html(content="ceci est un test", class_or_label_index=1)
        exp_list = explainer.explain_instance_as_list(content="ceci est un test", class_or_label_index=1)
        explanation = explainer.explain_instance(content="ceci est un test", class_or_label_index=2)
        html = explainer.explain_instance_as_html(content="ceci est un test", class_or_label_index=2)
        exp_list = explainer.explain_instance_as_list(content="ceci est un test", class_or_label_index=2)
        explanation = explainer.explain_instance(content="ceci est un test", class_or_label_index=None)
        html = explainer.explain_instance_as_html(content="ceci est un test", class_or_label_index=None)
        exp_list = explainer.explain_instance_as_list(content="ceci est un test", class_or_label_index=None)
        remove_dir(model_dir)

        # Check errors
        # No predict_proba
        with self.assertRaises(TypeError):
            LimeExplainer(5, model_conf)
        class FakeModelClass():
            def __init__(self):
                pass
            def predict_proba(self):
                pass
        # No list_classes
        with self.assertRaises(TypeError):
            LimeExplainer(FakeModelClass(), model_conf)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3

## Model embedding + LSTM + Attention
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - ModelEmbeddingLstm -> Model for predictions via embedding + LSTM


# ** EXPERIMENTAL **
# ** EXPERIMENTAL **
# ** EXPERIMENTAL **


import os
import json
import dill as pickle
import logging
import shutil
import numpy as np
import seaborn as sns
from typing import Union, Any, List, Callable

from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.preprocessing.text import Tokenizer
from tensorflow.keras.models import load_model as load_model_keras
from tensorflow.keras.layers import (Dense, Input, Embedding, GlobalMaxPooling1D,
                                     GlobalAveragePooling1D, BatchNormalization, LSTM,
                                     GRU, SpatialDropout1D, Bidirectional, concatenate)

from . import utils_deep_keras
from .model_keras import ModelKeras
from .utils_deep_keras import AttentionWithContext

sns.set(style="darkgrid")


class ModelEmbeddingLstmAttention(ModelKeras):
    '''Model for predictions via embedding + LSTM + Attention'''

    _default_name = 'model_embedding_lstm_attention'

    def __init__(self, max_sequence_length: int = 200, max_words: int = 100000,
                 padding: str = 'pre', truncating: str = 'post',
                 tokenizer_filters: str = "’!#$%&()*+,-./:;<=>?@[\\]^_`{|}~\t\n\r\'\"", **kwargs) -> None:
        '''Initialization of the class (see ModelClass & ModelKeras for more arguments)

        Kwargs:
            max_sequence_length (int): Maximum number of words per sequence (ie. sentences)
            max_words (int): Maximum number of words for tokenization
            padding (str): Padding (add zeros) at the beginning ('pre') or at the end ('post') of the sequences
            truncating (str): Truncating the beginning ('pre') or the end ('post') of the sequences (if superior to max_sequence_length)
            tokenizer_filters (str): Filter to use by the tokenizer
        Raises:
            ValueError: If the object padding is not a valid choice (['pre', 'post'])
            ValueError: If the object truncating is not a valid choice (['pre', 'post'])
        '''
        if padding not in ['pre', 'post']:
            raise ValueError(f"The object padding ({padding}) is not a valid choice (['pre', 'post'])")
        if truncating not in ['pre', 'post']:
            raise ValueError(f"The object truncating ({truncating}) is not a valid choice (['pre', 'post'])")
        # Init.
        super().__init__(**kwargs)

        # Get logger (must be done after super init)
        self.logger = logging.getLogger(__name__)

        self.max_sequence_length = max_sequence_length
        self.max_words = max_words
        self.padding = padding
        self.truncating = truncating

        # Tokenizer set on fit
        self.tokenizer: Any = None
        self.tokenizer_filters = tokenizer_filters

    def _prepare_x_train(self, x_train) -> np.ndarray:
        '''Prepares the input data for the model. Called when fitting the model

        Args:
            x_train (?): Array-like, shape = [n_samples, n_features]
        Returns:
            (np.ndarray): Prepared data
        '''
        # Get tokenizer & fit on train
        self.tokenizer = Tokenizer(num_words=self.max_words, filters=self.tokenizer_filters)
        self.logger.info('Fitting the tokenizer')
        self.tokenizer.fit_on_texts(x_train)
        return self._get_sequence(x_train, self.tokenizer, self.max_sequence_length, padding=self.padding, truncating=self.truncating)

    def _prepare_x_test(self, x_test) -> np.ndarray:
        '''Prepares the input data for the model. Called when fitting the model

        Args:
            x_test (?): Array-like, shape = [n_samples, n_features]
        Returns:
            (np.ndarray): Prepared data
        '''
        # Get sequences on test (already fitted on train)
        return self._get_sequence(x_test, self.tokenizer, self.max_sequence_length, padding=self.padding, truncating=self.truncating)

    def _get_model(self, custom_tokenizer=None) -> Any:
        '''Gets a model structure - returns the instance model instead if already defined

        Kwargs:
            custom_tokenizer (?): Tokenizer (if different from the one of the class). Permits to manage "new embeddings"
        Returns:
            (Model): a Keras model
        '''
        # Return model if already set
        if self.model is not None:
            return self.model

        # Start by getting embedding matrix
        if custom_tokenizer is not None:
            embedding_matrix, embedding_size = self._get_embedding_matrix(custom_tokenizer)
        else:
            embedding_matrix, embedding_size = self._get_embedding_matrix(self.tokenizer)

        # Get input dim
        input_dim = embedding_matrix.shape[0]

        # Get model
        num_classes = len(self.list_classes)
        # Process
        LSTM_UNITS = 100
        words = Input(shape=(self.max_sequence_length,))
        # trainable=True to finetune the model
        # words = Input(shape=(None,))
        # x = Embedding(*embedding_matrix.shape, weights=[embedding_matrix], trainable=False)(words)
        x = Embedding(input_dim, embedding_size, weights=[embedding_matrix], trainable=False)(words)
        x = BatchNormalization(momentum=0.9)(x)
        x = SpatialDropout1D(0.5)(x)
        # LSTM and GRU will default to CuDNNLSTM and CuDNNGRU if all conditions are met:
        # - activation = 'tanh'
        # - recurrent_activation = 'sigmoid'
        # - recurrent_dropout = 0
        # - unroll = False
        # - use_bias = True
        # - Inputs, if masked, are strictly right-padded
        # - reset_after = True (GRU only)
        # /!\ https://stackoverflow.com/questions/60468385/is-there-cudnnlstm-or-cudnngru-alternative-in-tensorflow-2-0
        x = Bidirectional(LSTM(LSTM_UNITS, return_sequences=True))(x)  # returns a sequence of vectors of dimension 32
        x = Bidirectional(GRU(LSTM_UNITS, return_sequences=True))(x)  # returns a sequence of vectors of dimension 32

        att = AttentionWithContext()(x)
        avg_pool1 = GlobalAveragePooling1D()(x)
        max_pool1 = GlobalMaxPooling1D()(x)

        x = concatenate([att, avg_pool1, max_pool1])
        # Last layer
        activation = 'sigmoid' if self.multi_label else 'softmax'
        out = Dense(num_classes, activation=activation, kernel_initializer='glorot_uniform')(x)

        # Compile model
        model = Model(inputs=words, outputs=[out])
        lr = self.keras_params['learning_rate'] if 'learning_rate' in self.keras_params.keys() else 0.001
        decay = self.keras_params['decay'] if 'decay' in self.keras_params.keys() else 0.0
        self.logger.info(f"Learning rate: {lr}")
        self.logger.info(f"Decay: {decay}")
        optimizer = Adam(lr=lr, decay=decay)
        loss = utils_deep_keras.f1_loss if self.multi_label else 'categorical_crossentropy'
        # loss = 'binary_crossentropy' if self.multi_label else 'categorical_crossentropy'  # utils_deep_keras.f1_loss also possible if multi-labels
        metrics: List[Union[str, Callable]] = ['accuracy'] if not self.multi_label else ['categorical_accuracy', utils_deep_keras.f1, utils_deep_keras.precision, utils_deep_keras.recall]
        model.compile(optimizer=optimizer, loss=loss, metrics=metrics)
        if self.logger.getEffectiveLevel() < logging.ERROR:
            model.summary()

        # Try to save model as png if level_save > 'LOW'
        if self.level_save in ['MEDIUM', 'HIGH']:
            self._save_model_png(model)

        # Return
        return model

    def save(self, json_data: Union[dict, None] = None) -> None:
        '''Saves the model

        Kwargs:
            json_data (dict): Additional configurations to be saved
        '''
        # Save configuration JSON
        if json_data is None:
            json_data = {}

        # Add specific data
        json_data['max_sequence_length'] = self.max_sequence_length
        json_data['max_words'] = self.max_words
        json_data['padding'] = self.padding
        json_data['truncating'] = self.truncating
        json_data['tokenizer_filters'] = self.tokenizer_filters

        # Save tokenizer if not None & level_save > LOW
        if (self.tokenizer is not None) and (self.level_save in ['MEDIUM', 'HIGH']):
            # Manage paths
            tokenizer_path = os.path.join(self.model_dir, "embedding_tokenizer.pkl")
            # Save as pickle
            with open(tokenizer_path, 'wb') as f:
                pickle.dump(self.tokenizer, f)

        # Save
        super().save(json_data=json_data)

    @classmethod
    def _init_new_instance_from_configs(cls, configs):
        '''Inits a new instance from a set of configurations

        Args:
            configs: a set of configurations of a model to be reloaded
        Returns:
            ModelClass: the newly generated class
        '''
        # Call parent
        model = super()._init_new_instance_from_configs(configs)

        # Try to read the following attributes from configs and, if absent, keep the current one
        for attribute in ['max_sequence_length', 'max_words', 'padding', 'truncating', 'tokenizer_filters']:
            setattr(model, attribute, configs.get(attribute, getattr(model, attribute)))

        # Return the new model
        return model

    def _load_standalone_files(self, default_model_dir: Union[str, None] = None,  # type: ignore
                               tokenizer_path: Union[str, None] = None, *args, **kwargs):
        '''Loads standalone files for a newly created model via _init_new_instance_from_configs

        Kwargs:
            default_model_dir (str): a path to look for default file paths
                                     If None, standalone files path should all be provided
            tokenizer_path (str): Path to the tokenizer file
        Raises:
            ValueError: If the tokenizer file is not specified and can't be inferred
            FileNotFoundError: If the tokenizer file does not exist
        '''
        # Check if we are able to get all needed paths
        if default_model_dir is None and tokenizer_path is None:
            raise ValueError("The tokenizer file is not specified and can't be inferred")

        # Call parent
        super()._load_standalone_files(default_model_dir=default_model_dir, **kwargs)

        # Retrieve file paths
        if tokenizer_path is None:
            tokenizer_path = os.path.join(default_model_dir, "embedding_tokenizer.pkl")

        # Check paths exists
        if not os.path.isfile(tokenizer_path):
            raise FileNotFoundError(f"Can't find tokenizer file ({tokenizer_path})")

        # Reload tokenizer
        with open(tokenizer_path, 'rb') as f:
            self.tokenizer = pickle.load(f)


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import io
import cv2
import copy
import math
import json
import shutil
import numpy as np
import pandas as pd
from PIL import Image

from {{package_name}} import utils
from {{package_name}}.preprocessing import preprocess
from {{package_name}}.models_training.object_detectors import utils_faster_rcnn
from {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn import (ModelKerasFasterRcnnObjectDetector,
                                                                                       ModelCheckpointAll,
                                                                                       CustomGeneratorRpn,
                                                                                       CustomGeneratorClassifier)

import tensorflow as tf
from tensorflow.keras.layers import Input


# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)
    tf.keras.backend.clear_session()


# Mocks

def download_url_crash(x, y):
    raise ConnectionError("error")


class MockCustomGeneratorRpn():
    def __init__(self, img_data_list: list, batch_size: int, shuffle: bool, seed: int, model, **kwargs) -> None:
        self.n = len(img_data_list)
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.seed = seed
        self.img_data_list = img_data_list
        self.model = model
        # Default arguments
        self.horizontal_flip = kwargs.get('horizontal_flip', False)
        self.vertical_flip = kwargs.get('vertical_flip', False)
        self.rot_90 = kwargs.get('rot_90', False)
        self.data_type = kwargs.get('data_type', 'train')
        self.with_img_data = kwargs.get('with_img_data', False)

class MockCustomGeneratorClassifier(MockCustomGeneratorRpn):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # Default arguments
        self.shared_model_trainable = kwargs.get('shared_model_trainable', False)

class MockModelKerasFasterRcnnObjectDetector():
    def __init__(self) -> None:
        self.anchor_box_sizes = [64, 128, 256]  # In the paper : [128, 256, 512]
        self.anchor_box_ratios = [[1, 1], [1./math.sqrt(2), 2./math.sqrt(2)], [2./math.sqrt(2), 1./math.sqrt(2)]]  # In the paper : [1, 1], [1, 2], [2, 1]]
        self.nb_anchors = len(self.anchor_box_sizes) * len(self.anchor_box_ratios)
        self.list_anchors = [[anchor_size * anchor_ratio[0], anchor_size * anchor_ratio[1]]
                             for anchor_size in self.anchor_box_sizes for anchor_ratio in self.anchor_box_ratios]
        self.shared_model_subsampling = 16
        self.rpn_min_overlap = 0.2
        self.rpn_max_overlap = 0.8
        self.rpn_regr_scaling = 4
        self.rpn_restrict_num_regions = 100
        self.logger = logging.getLogger(__name__)
        self.preprocess_input = self._get_preprocess_input()
        # Manage RPN model
        class FakeModelRpn():
            def __init__(self) -> None:
                pass
            def set_weights(self, model):
                pass
            def get_weights(self):
                return None
        self.model_rpn = FakeModelRpn()
    def _get_preprocess_input(self):
        def preprocess_input(img):
            return img
        return preprocess_input
    def _generate_images_with_bboxes(self, img_data: dict, horizontal_flip: bool = False,
                                     vertical_flip: bool = False, rot_90: bool = False):
        with open(img_data['file_path'], 'rb') as f:
            img = Image.open(io.BytesIO(f.read()))
            if img.mode != 'RGB':
                img = img.convert('RGB')
            img = np.asarray(img)
        bboxes = copy.deepcopy(img_data.get('bboxes', []))  # Empty if test
        original_height, original_width = img.shape[0], img.shape[1]
        img = self.preprocess_input(img)
        resized_height, resized_width = img.shape[0], img.shape[1]
        for bbox in bboxes:
            bbox['x1'] = bbox['x1'] * (resized_width / original_width)
            bbox['x2'] = bbox['x2'] * (resized_width / original_width)
            bbox['y1'] = bbox['y1'] * (resized_height / original_height)
            bbox['y2'] = bbox['y2'] * (resized_height / original_height)
        prepared_data = {'img': img, 'bboxes': bboxes, 'original_height': original_height,
                         'original_width': original_width, 'resized_height': resized_height,
                         'resized_width': resized_width}
        return prepared_data

def mock_get_rpn_targets(model, img_data_batch: list):
    return 'toto', 'titi'

def mock_clone_model_fn(model, input_tensors=None, clone_function=None):
    return model

# Tests

class ModelKerasFasterRcnnObjectDetectorTests(unittest.TestCase):
    '''Main class to test model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector'''


    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test01_model_keras_faster_rcnn_init(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector.__init__'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Init., test all parameters
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        self.assertEqual(model.model_dir, model_dir)
        self.assertEqual(model.model_type, 'object_detector')
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.model, None)
        self.assertEqual(model.shared_model, None)
        self.assertEqual(model.model_rpn, None)
        self.assertEqual(model.model_classifier, None)
        # We test display_if_gpu_activated and _is_gpu_activated just by calling them
        self.assertTrue(type(model._is_gpu_activated()) == bool)
        model.display_if_gpu_activated()
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, epochs=8)
        self.assertEqual(model.epochs, 8)
        self.assertEqual(model.epochs_rpn_trainable_true, 8)
        self.assertEqual(model.epochs_classifier_trainable_true, 8)
        self.assertEqual(model.epochs_rpn_trainable_false, 8)
        self.assertEqual(model.epochs_classifier_trainable_false, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, epochs=8, epochs_rpn_trainable_true=2,
                                                   epochs_classifier_trainable_true=3, epochs_rpn_trainable_false=4,
                                                   epochs_classifier_trainable_false=5)
        self.assertEqual(model.epochs, 8)
        self.assertEqual(model.epochs_rpn_trainable_true, 2)
        self.assertEqual(model.epochs_classifier_trainable_true, 3)
        self.assertEqual(model.epochs_rpn_trainable_false, 4)
        self.assertEqual(model.epochs_classifier_trainable_false, 5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, batch_size=8)
        self.assertEqual(model.batch_size, 8)
        self.assertEqual(model.batch_size_rpn_trainable_true, 8)
        self.assertEqual(model.batch_size_classifier_trainable_true, 8)
        self.assertEqual(model.batch_size_rpn_trainable_false, 8)
        self.assertEqual(model.batch_size_classifier_trainable_false, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, batch_size=8, batch_size_rpn_trainable_true=2,
                                                   batch_size_classifier_trainable_true=3, batch_size_rpn_trainable_false=4,
                                                   batch_size_classifier_trainable_false=5)
        self.assertEqual(model.batch_size, 8)
        self.assertEqual(model.batch_size_rpn_trainable_true, 2)
        self.assertEqual(model.batch_size_classifier_trainable_true, 3)
        self.assertEqual(model.batch_size_rpn_trainable_false, 4)
        self.assertEqual(model.batch_size_classifier_trainable_false, 5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, patience=8)
        self.assertEqual(model.patience, 8)
        self.assertEqual(model.patience_rpn_trainable_true, 8)
        self.assertEqual(model.patience_classifier_trainable_true, 8)
        self.assertEqual(model.patience_rpn_trainable_false, 8)
        self.assertEqual(model.patience_classifier_trainable_false, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, patience=8, patience_rpn_trainable_true=2,
                                                   patience_classifier_trainable_true=3, patience_rpn_trainable_false=4,
                                                   patience_classifier_trainable_false=5)
        self.assertEqual(model.patience, 8)
        self.assertEqual(model.patience_rpn_trainable_true, 2)
        self.assertEqual(model.patience_classifier_trainable_true, 3)
        self.assertEqual(model.patience_rpn_trainable_false, 4)
        self.assertEqual(model.patience_classifier_trainable_false, 5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, lr_rpn_trainable_true=0.2, lr_classifier_trainable_true=0.3,
                                                   lr_rpn_trainable_false=0.4, lr_classifier_trainable_false=0.5)
        self.assertEqual(model.keras_params['lr_rpn_trainable_true'], 0.2)
        self.assertEqual(model.keras_params['lr_classifier_trainable_true'], 0.3)
        self.assertEqual(model.keras_params['lr_rpn_trainable_false'], 0.4)
        self.assertEqual(model.keras_params['lr_classifier_trainable_false'], 0.5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=64)
        self.assertEqual(model.img_min_side_size, 64)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=0.8, rpn_max_overlap=0.9)
        self.assertEqual(model.rpn_min_overlap, 0.8)
        self.assertEqual(model.rpn_max_overlap, 0.9)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_restrict_num_regions=8)
        self.assertEqual(model.rpn_restrict_num_regions, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pool_resize_classifier=8)
        self.assertEqual(model.pool_resize_classifier, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nb_rois_classifier=8)
        self.assertEqual(model.nb_rois_classifier, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, roi_nms_overlap_threshold=0.8)
        self.assertEqual(model.roi_nms_overlap_threshold, 0.8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nms_max_boxes=8)
        self.assertEqual(model.nms_max_boxes, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=0.8, classifier_max_overlap=0.9)
        self.assertEqual(model.classifier_min_overlap, 0.8)
        self.assertEqual(model.classifier_max_overlap, 0.9)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_bbox_proba_threshold=0.8)
        self.assertEqual(model.pred_bbox_proba_threshold, 0.8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_nms_overlap_threshold=0.8)
        self.assertEqual(model.pred_nms_overlap_threshold, 0.8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'toto': 'titi'})
        self.assertEqual(model.data_augmentation_params, {'toto': 'titi'})
        remove_dir(model_dir)

        # Check errors
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=31)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_max_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_max_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=0.3, rpn_max_overlap=0.1)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_restrict_num_regions=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pool_resize_classifier=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nb_rois_classifier=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, roi_nms_overlap_threshold=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, roi_nms_overlap_threshold=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nms_max_boxes=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_max_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_max_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=0.3, classifier_max_overlap=0.1)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_bbox_proba_threshold=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_bbox_proba_threshold=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_nms_overlap_threshold=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_nms_overlap_threshold=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, color_mode='rgba')
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, color_mode='grayscale')

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    @patch('{{package_name}}.utils.download_url', side_effect=download_url_crash)
    def test02_model_keras_faster_rcnn_init_offline(self, mock_download_url):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector.__init__
        - No access to a base model
        '''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Clean cache path if exists
        cache_path = os.path.join(utils.get_data_path(), 'transfer_learning_weights')
        if os.path.isdir(cache_path):
            shutil.rmtree(cache_path)

        # Init., test all parameters
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        self.assertEqual(model.model_dir, model_dir)
        self.assertEqual(model.model_type, 'object_detector')
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.model, None)
        self.assertEqual(model.shared_model, None)
        self.assertEqual(model.model_rpn, None)
        self.assertEqual(model.model_classifier, None)
        # We test display_if_gpu_activated and _is_gpu_activated just by calling them
        self.assertTrue(type(model._is_gpu_activated()) == bool)
        model.display_if_gpu_activated()
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, epochs=8)
        self.assertEqual(model.epochs, 8)
        self.assertEqual(model.epochs_rpn_trainable_true, 8)
        self.assertEqual(model.epochs_classifier_trainable_true, 8)
        self.assertEqual(model.epochs_rpn_trainable_false, 8)
        self.assertEqual(model.epochs_classifier_trainable_false, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, epochs=8, epochs_rpn_trainable_true=2,
                                                   epochs_classifier_trainable_true=3, epochs_rpn_trainable_false=4,
                                                   epochs_classifier_trainable_false=5)
        self.assertEqual(model.epochs, 8)
        self.assertEqual(model.epochs_rpn_trainable_true, 2)
        self.assertEqual(model.epochs_classifier_trainable_true, 3)
        self.assertEqual(model.epochs_rpn_trainable_false, 4)
        self.assertEqual(model.epochs_classifier_trainable_false, 5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, batch_size=8)
        self.assertEqual(model.batch_size, 8)
        self.assertEqual(model.batch_size_rpn_trainable_true, 8)
        self.assertEqual(model.batch_size_classifier_trainable_true, 8)
        self.assertEqual(model.batch_size_rpn_trainable_false, 8)
        self.assertEqual(model.batch_size_classifier_trainable_false, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, batch_size=8, batch_size_rpn_trainable_true=2,
                                                   batch_size_classifier_trainable_true=3, batch_size_rpn_trainable_false=4,
                                                   batch_size_classifier_trainable_false=5)
        self.assertEqual(model.batch_size, 8)
        self.assertEqual(model.batch_size_rpn_trainable_true, 2)
        self.assertEqual(model.batch_size_classifier_trainable_true, 3)
        self.assertEqual(model.batch_size_rpn_trainable_false, 4)
        self.assertEqual(model.batch_size_classifier_trainable_false, 5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, patience=8)
        self.assertEqual(model.patience, 8)
        self.assertEqual(model.patience_rpn_trainable_true, 8)
        self.assertEqual(model.patience_classifier_trainable_true, 8)
        self.assertEqual(model.patience_rpn_trainable_false, 8)
        self.assertEqual(model.patience_classifier_trainable_false, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, patience=8, patience_rpn_trainable_true=2,
                                                   patience_classifier_trainable_true=3, patience_rpn_trainable_false=4,
                                                   patience_classifier_trainable_false=5)
        self.assertEqual(model.patience, 8)
        self.assertEqual(model.patience_rpn_trainable_true, 2)
        self.assertEqual(model.patience_classifier_trainable_true, 3)
        self.assertEqual(model.patience_rpn_trainable_false, 4)
        self.assertEqual(model.patience_classifier_trainable_false, 5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, lr_rpn_trainable_true=0.2, lr_classifier_trainable_true=0.3,
                                                   lr_rpn_trainable_false=0.4, lr_classifier_trainable_false=0.5)
        self.assertEqual(model.keras_params['lr_rpn_trainable_true'], 0.2)
        self.assertEqual(model.keras_params['lr_classifier_trainable_true'], 0.3)
        self.assertEqual(model.keras_params['lr_rpn_trainable_false'], 0.4)
        self.assertEqual(model.keras_params['lr_classifier_trainable_false'], 0.5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=64)
        self.assertEqual(model.img_min_side_size, 64)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=0.8, rpn_max_overlap=0.9)
        self.assertEqual(model.rpn_min_overlap, 0.8)
        self.assertEqual(model.rpn_max_overlap, 0.9)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_restrict_num_regions=8)
        self.assertEqual(model.rpn_restrict_num_regions, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pool_resize_classifier=8)
        self.assertEqual(model.pool_resize_classifier, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nb_rois_classifier=8)
        self.assertEqual(model.nb_rois_classifier, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, roi_nms_overlap_threshold=0.8)
        self.assertEqual(model.roi_nms_overlap_threshold, 0.8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nms_max_boxes=8)
        self.assertEqual(model.nms_max_boxes, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=0.8, classifier_max_overlap=0.9)
        self.assertEqual(model.classifier_min_overlap, 0.8)
        self.assertEqual(model.classifier_max_overlap, 0.9)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_bbox_proba_threshold=0.8)
        self.assertEqual(model.pred_bbox_proba_threshold, 0.8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_nms_overlap_threshold=0.8)
        self.assertEqual(model.pred_nms_overlap_threshold, 0.8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'toto': 'titi'})
        self.assertEqual(model.data_augmentation_params, {'toto': 'titi'})
        remove_dir(model_dir)

        # Check errors
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=31)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_max_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_max_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=0.3, rpn_max_overlap=0.1)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_restrict_num_regions=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pool_resize_classifier=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nb_rois_classifier=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, roi_nms_overlap_threshold=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, roi_nms_overlap_threshold=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nms_max_boxes=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_max_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_max_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=0.3, classifier_max_overlap=0.1)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_bbox_proba_threshold=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_bbox_proba_threshold=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_nms_overlap_threshold=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_nms_overlap_threshold=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, color_mode='rgba')
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, color_mode='grayscale')

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test03_model_keras_faster_rcnn_get_model(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._get_model'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        model.list_classes = ['orange', 'banana', 'apple']
        model.dict_classes = {i: col for i, col in enumerate(model.list_classes)}
        model.custom_objects = {**utils_faster_rcnn.get_custom_objects_faster_rcnn(model.nb_anchors, len(model.list_classes)), **model.custom_objects}
        model._get_model()  # Nothing to test in particular ...

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test04_model_keras_faster_rcnn_compile_model_rpn(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._compile_model_rpn'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, lr_rpn_trainable_true=0.1)
        model.list_classes = ['orange', 'banana', 'apple']
        model.dict_classes = {i: col for i, col in enumerate(model.list_classes)}
        model.custom_objects = {**utils_faster_rcnn.get_custom_objects_faster_rcnn(model.nb_anchors, len(model.list_classes)), **model.custom_objects}
        model.shared_model, model.model_rpn, model.model_classifier, model.model = model._get_model()
        self.assertAlmostEqual(0.1, float(model.model_rpn.optimizer.lr.numpy()))
        model._compile_model_rpn(model.model_rpn, 0.1234)
        self.assertAlmostEqual(0.1234, float(model.model_rpn.optimizer.lr.numpy()))

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test05_model_keras_faster_rcnn_compile_model_classifier(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._compile_model_classifier'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, lr_classifier_trainable_true=0.1)
        model.list_classes = ['orange', 'banana', 'apple']
        model.dict_classes = {i: col for i, col in enumerate(model.list_classes)}
        model.custom_objects = {**utils_faster_rcnn.get_custom_objects_faster_rcnn(model.nb_anchors, len(model.list_classes)), **model.custom_objects}
        model.shared_model, model.model_rpn, model.model_classifier, model.model = model._get_model()
        self.assertAlmostEqual(0.1, float(model.model_classifier.optimizer.lr.numpy()))
        model._compile_model_classifier(model.model_classifier, 0.1234)
        self.assertAlmostEqual(0.1234, float(model.model_classifier.optimizer.lr.numpy()))

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test06_model_keras_faster_rcnn_get_shared_model_structure(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._get_shared_model_structure'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        input_img = Input(shape=(None, None, 3), name='input_img')
        model._get_shared_model_structure(input_img)  # Nothing to test in particular ...

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test07_model_keras_faster_rcnn_add_rpn_layers(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._add_rpn_layers'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        input_img = Input(shape=(None, None, 3), name='input_img')
        shared_model_layers = model._get_shared_model_structure(input_img)
        model._add_rpn_layers(shared_model_layers)  # Nothing to test in particular ...

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test08_model_keras_faster_rcnn_add_classifier_layers(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._add_classifier_layers'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        model.list_classes = ['orange', 'banana', 'apple']
        model.dict_classes = {i: col for i, col in enumerate(model.list_classes)}
        model.custom_objects = {**utils_faster_rcnn.get_custom_objects_faster_rcnn(model.nb_anchors, len(model.list_classes)), **model.custom_objects}
        input_img = Input(shape=(None, None, 3), name='input_img')
        input_rois = Input(shape=(None, model.nb_rois_classifier), name='input_rois')
        shared_model_layers = model._get_shared_model_structure(input_img)
        rpn_layers = model._add_rpn_layers(shared_model_layers)
        model._add_classifier_layers(shared_model_layers, input_rois)  # Nothing to test in particular ...

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    @patch('{{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.CustomGeneratorClassifier', side_effect=MockCustomGeneratorClassifier)
    @patch('{{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.CustomGeneratorRpn', side_effect=MockCustomGeneratorRpn)
    def test09_model_keras_faster_rcnn_get_generator(self, mock_custom_generator_rpn, mock_custom_generator_classifier):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._get_generator'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars (no need to put the real data, the generator is mocked)
        df = pd.DataFrame({
            'file_path': ['a', 'c', 'e', 'b'],
            'bboxes': [[[1, 2], [4, 2]], [[1, 5]], [[3, 8]], [[5, 2], [4, 3]]],
        })

        # Nominal case - RPN
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df, data_type='train', batch_size=5, generator_type='rpn', with_img_data=False)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, True)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, True)
        self.assertEqual(generator.vertical_flip, True)
        self.assertEqual(generator.rot_90, True)
        self.assertEqual(generator.data_type, 'train')
        self.assertEqual(generator.with_img_data, False)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': False, 'vertical_flip': False, 'rot_90': False})
        generator = model._get_generator(df, data_type='train', batch_size=5, generator_type='rpn', with_img_data=False)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, True)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'train')
        self.assertEqual(generator.with_img_data, False)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df, data_type='valid', batch_size=10, generator_type='rpn', with_img_data=True)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 10)
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'valid')
        self.assertEqual(generator.with_img_data, True)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df, data_type='test', batch_size=5, generator_type='rpn', with_img_data=False)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'test')
        self.assertEqual(generator.with_img_data, False)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df[['file_path']], data_type='test', batch_size=5, generator_type='rpn', with_img_data=False)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'test')
        self.assertEqual(generator.with_img_data, False)
        remove_dir(model_dir)

        # Nominal case - Classifier
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df, data_type='train', batch_size=5, generator_type='classifier', with_img_data=False, shared_model_trainable=True)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, True)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, True)
        self.assertEqual(generator.vertical_flip, True)
        self.assertEqual(generator.rot_90, True)
        self.assertEqual(generator.data_type, 'train')
        self.assertEqual(generator.with_img_data, False)
        self.assertEqual(generator.shared_model_trainable, True)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': False, 'vertical_flip': False, 'rot_90': False})
        generator = model._get_generator(df, data_type='train', batch_size=5, generator_type='classifier', with_img_data=False, shared_model_trainable=True)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, True)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'train')
        self.assertEqual(generator.with_img_data, False)
        self.assertEqual(generator.shared_model_trainable, True)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df, data_type='valid', batch_size=10, generator_type='classifier', with_img_data=True, shared_model_trainable=False)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 10)
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'valid')
        self.assertEqual(generator.with_img_data, True)
        self.assertEqual(generator.shared_model_trainable, False)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df, data_type='test', batch_size=5, generator_type='classifier', with_img_data=False, shared_model_trainable=True)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'test')
        self.assertEqual(generator.with_img_data, False)
        self.assertEqual(generator.shared_model_trainable, True)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df[['file_path']], data_type='test', batch_size=5, generator_type='classifier', with_img_data=False, shared_model_trainable=True)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'test')
        self.assertEqual(generator.with_img_data, False)
        self.assertEqual(generator.shared_model_trainable, True)
        remove_dir(model_dir)

        # Manage errors
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        with self.assertRaises(ValueError):
            generator = model._get_generator(df, data_type='toto', batch_size=5, generator_type='rpn')
        with self.assertRaises(ValueError):
            generator = model._get_generator(df, data_type='train', batch_size=5, generator_type='toto')
        with self.assertRaises(ValueError):
            generator = model._get_generator(pd.DataFrame(columns=['toto']), data_type='train', batch_size=5, generator_type='rpn')
        with self.assertRaises(ValueError):
            generator = model._get_generator(pd.DataFrame(columns=['file_path']), data_type='train', batch_size=5, generator_type='rpn')
        with self.assertRaises(ValueError):
            generator = model._get_generator(pd.DataFrame(columns=['file_path']), data_type='valid', batch_size=5, generator_type='rpn')
        with self.assertRaises(ValueError):
            model.model_type = 'toto'
            generator = model._get_generator(df, data_type='train', batch_size=5, generator_type='rpn')

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    @patch('numpy.random.choice')
    def test10_model_keras_faster_rcnn_generate_images_with_bboxes(self, mock_np_random_choice):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._generate_images_with_bboxes'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        im_path = os.path.join(os.getcwd(), 'test_data', 'apple_36.jpg')
        img_data = {'file_path': im_path, 'bboxes': [{'x1': 10, 'x2': 30, 'y1': 10, 'y2': 50}, {'x1': 5, 'x2': 25, 'y1': 10, 'y2': 45}]}
        img_data_simple = {'file_path': im_path}
        with Image.open(im_path) as im:
            w, h = im.size
        resize_w, resize_h = (250, 300)
        # Mock preprocess
        def fake_preprocess(img):
            img = cv2.resize(img, (resize_w, resize_h), interpolation=cv2.INTER_CUBIC)
            return img

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        model.preprocess_input = fake_preprocess
        prepared_data = model._generate_images_with_bboxes(img_data, horizontal_flip=False, vertical_flip=False, rot_90=False)
        self.assertEqual(prepared_data['img'].shape[0], resize_h)
        self.assertEqual(prepared_data['img'].shape[1], resize_w)
        self.assertEqual(len(prepared_data['bboxes']), len(img_data['bboxes']))
        self.assertNotEqual(prepared_data['bboxes'][0]['x1'], img_data['bboxes'][0]['x1'])
        self.assertNotEqual(prepared_data['bboxes'][0]['y1'], img_data['bboxes'][0]['y1'])
        self.assertEqual(prepared_data['original_height'], h)
        self.assertEqual(prepared_data['original_width'], w)
        self.assertEqual(prepared_data['resized_height'], resize_h)
        self.assertEqual(prepared_data['resized_width'], resize_w)
        # Without bbox
        prepared_data = model._generate_images_with_bboxes(img_data_simple, horizontal_flip=False, vertical_flip=False, rot_90=False)
        self.assertEqual(prepared_data['img'].shape[0], resize_h)
        self.assertEqual(prepared_data['img'].shape[1], resize_w)
        self.assertEqual(prepared_data['bboxes'], [])
        self.assertEqual(prepared_data['original_height'], h)
        self.assertEqual(prepared_data['original_width'], w)
        self.assertEqual(prepared_data['resized_height'], resize_h)
        self.assertEqual(prepared_data['resized_width'], resize_w)
        remove_dir(model_dir)

        # Tests augmentation
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        model.preprocess_input = fake_preprocess
        prepared_data = model._generate_images_with_bboxes(img_data, horizontal_flip=True, vertical_flip=False, rot_90=False)
        self.assertEqual(prepared_data['img'].shape[0], resize_h)
        self.assertEqual(prepared_data['img'].shape[1], resize_w)
        self.assertEqual(len(prepared_data['bboxes']), len(img_data['bboxes']))
        self.assertNotEqual(prepared_data['bboxes'][0]['x1'], img_data['bboxes'][0]['x1'])
        self.assertNotEqual(prepared_data['bboxes'][0]['y1'], img_data['bboxes'][0]['y1'])
        self.assertEqual(prepared_data['original_height'], h)
        self.assertEqual(prepared_data['original_width'], w)
        self.assertEqual(prepared_data['resized_height'], resize_h)
        self.assertEqual(prepared_data['resized_width'], resize_w)
        #
        prepared_data = model._generate_images_with_bboxes(img_data_simple, horizontal_flip=False, vertical_flip=True, rot_90=False)
        self.assertEqual(prepared_data['img'].shape[0], resize_h)
        self.assertEqual(prepared_data['img'].shape[1], resize_w)
        self.assertEqual(prepared_data['bboxes'], [])
        self.assertEqual(prepared_data['original_height'], h)
        self.assertEqual(prepared_data['original_width'], w)
        self.assertEqual(prepared_data['resized_height'], resize_h)
        self.assertEqual(prepared_data['resized_width'], resize_w)
        remove_dir(model_dir)
        #
        mock_np_random_choice.return_value = [90]
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        model.preprocess_input = fake_preprocess
        prepared_data = model._generate_images_with_bboxes(img_data, horizontal_flip=False, vertical_flip=False, rot_90=True)
        self.assertEqual(prepared_data['img'].shape[0], resize_h)
        self.assertEqual(prepared_data['img'].shape[1], resize_w)
        self.assertEqual(len(prepared_data['bboxes']), len(img_data['bboxes']))
        self.assertNotEqual(prepared_data['bboxes'][0]['x1'], img_data['bboxes'][0]['x1'])
        self.assertNotEqual(prepared_data['bboxes'][0]['y1'], img_data['bboxes'][0]['y1'])
        self.assertEqual(prepared_data['original_height'], w)  # Rot 90
        self.assertEqual(prepared_data['original_width'], h)  # Rot 90
        self.assertEqual(prepared_data['resized_height'], resize_h)
        self.assertEqual(prepared_data['resized_width'], resize_w)
        #
        mock_np_random_choice.return_value = [180]
        prepared_data = model._generate_images_with_bboxes(img_data_simple, horizontal_flip=False, vertical_flip=False, rot_90=True)
        self.assertEqual(prepared_data['img'].shape[0], resize_h)
        self.assertEqual(prepared_data['img'].shape[1], resize_w)
        self.assertEqual(prepared_data['bboxes'], [])
        self.assertEqual(prepared_data['original_height'], h)  # Rot 180
        self.assertEqual(prepared_data['original_width'], w)  # Rot 180
        self.assertEqual(prepared_data['resized_height'], resize_h)
        self.assertEqual(prepared_data['resized_width'], resize_w)
        remove_dir(model_dir)

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test11_model_keras_faster_rcnn_get_preprocess_input(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._get_preprocess_input'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        im_path = os.path.join(os.getcwd(), 'test_data', 'apple_36.jpg')
        with open(im_path, 'rb') as f:
            img = Image.open(io.BytesIO(f.read()))
            if img.mode != 'RGB':
                img = img.convert('RGB')
            # Convert to array
            img = np.asarray(img)
        h, w = img.shape[:2]  # H > W

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100)
        preprocess_input = model._get_preprocess_input()
        new_img = preprocess_input(img)
        new_h, new_w = new_img.shape[:2]
        self.assertEqual(new_w, 100)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=350)
        preprocess_input = model._get_preprocess_input()
        new_img = preprocess_input(img)
        new_h, new_w = new_img.shape[:2]
        self.assertEqual(new_w, 350)

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test12_model_keras_faster_rcnn_fit_object_detector(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._fit_object_detector'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        df_data = pd.DataFrame({'file_path': path_list, 'bboxes': bboxes_list}).copy()

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10,
                                                   lr_rpn_trainable_true=0.011, lr_classifier_trainable_true=0.012, lr_rpn_trainable_false=0.013,
                                                   lr_classifier_trainable_false=0.014, keras_params={'decay_rpn': 0.0, 'decay_classifier': 0.0})
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model._fit_object_detector(df_data, df_valid=None, with_shuffle=True)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(sorted(model.list_classes), ['apple', 'banana', 'orange'])
        self.assertTrue(model.model_rpn._is_compiled)
        self.assertTrue(model.model_classifier._is_compiled)
        self.assertTrue(model.model._is_compiled)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5')))
        self.assertAlmostEqual(0.013, float(model.model_rpn.optimizer._decayed_lr(tf.float32).numpy()))
        self.assertAlmostEqual(0.014, float(model.model_classifier.optimizer._decayed_lr(tf.float32).numpy()))
        remove_dir(model_dir)
        # With valid & shuffle à False
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10,
                                                   lr_rpn_trainable_true=0.011, lr_classifier_trainable_true=0.012, lr_rpn_trainable_false=0.013,
                                                   lr_classifier_trainable_false=0.014, keras_params={'decay_rpn': 0.0, 'decay_classifier': 0.0})
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model._fit_object_detector(df_data, df_valid=df_data, with_shuffle=False)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(sorted(model.list_classes), ['apple', 'banana', 'orange'])
        self.assertTrue(model.model_rpn._is_compiled)
        self.assertTrue(model.model_classifier._is_compiled)
        self.assertTrue(model.model._is_compiled)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5')))
        self.assertAlmostEqual(0.013, float(model.model_rpn.optimizer._decayed_lr(tf.float32).numpy()))
        self.assertAlmostEqual(0.014, float(model.model_classifier.optimizer._decayed_lr(tf.float32).numpy()))
        remove_dir(model_dir)

        # Test continue training
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10,
                                                   lr_rpn_trainable_true=0.011, lr_classifier_trainable_true=0.012, lr_rpn_trainable_false=0.013,
                                                   lr_classifier_trainable_false=0.014, keras_params={'decay_rpn': 0.0, 'decay_classifier': 0.0})
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(df_data, df_valid=df_data, with_shuffle=True)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        model.save()
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        # 2nd fit
        model.fit(df_data, df_valid=df_data, with_shuffle=True)  # We fit again with the same data, not important
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 2)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_1.json')))
        # We do not save on purpose
        model_dir_2 = model.model_dir
        self.assertNotEqual(model_dir, model_dir_2)
        # 3rd fit
        model.fit(df_data, df_valid=df_data, with_shuffle=True)  # We fit again with the same data, not important
        model.save()
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 3)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_1.json')))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_2.json')))
        model_dir_3 = model.model_dir
        self.assertNotEqual(model_dir_2, model_dir_3)
        # Assertion errors (bad classes)
        old_dict_classes = model.dict_classes.copy()
        model.dict_classes = {'a': 0, 'b': 1}
        with self.assertRaises(ValueError):
            model.fit(df_data, df_valid=df_data, with_shuffle=True)
        model.dict_classes = old_dict_classes
        model.list_classes = ['a', 'b']
        with self.assertRaises(ValueError):
            model.fit(df_data, df_valid=df_data, with_shuffle=True)
        remove_dir(model_dir)
        remove_dir(model_dir_2)
        remove_dir(model_dir_3)

        # Manage errors
        df_data['bboxes'][0][0]['class'] = 'bg'
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10,
                                                   lr_rpn_trainable_true=0.011, lr_classifier_trainable_true=0.012, lr_rpn_trainable_false=0.013,
                                                   lr_classifier_trainable_false=0.014, keras_params={'decay_rpn': 0.0, 'decay_classifier': 0.0})
        with self.assertRaises(ValueError):
            model.fit(df_data, df_valid=df_data, with_shuffle=True)

        # Clean
        remove_dir(model_dir)


    # We do not test _fit_object_detector_RPN & _fit_object_detector_classifier, already done via _fit_object_detector

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test13_model_keras_faster_rcnn_get_callbacks(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._get_callbacks'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        callbacks = model._get_callbacks(patience=12)
        callbacks_types = [type(_) for _ in callbacks]
        self.assertTrue(tf.keras.callbacks.EarlyStopping in callbacks_types)
        self.assertTrue(ModelCheckpointAll in callbacks_types)
        self.assertTrue(tf.keras.callbacks.CSVLogger in callbacks_types)
        self.assertTrue(tf.keras.callbacks.TerminateOnNaN in callbacks_types)
        checkpoint = callbacks[callbacks_types.index(ModelCheckpointAll)]
        csv_logger = callbacks[callbacks_types.index(tf.keras.callbacks.CSVLogger)]
        early_stopping = callbacks[callbacks_types.index(tf.keras.callbacks.EarlyStopping)]
        self.assertEqual(checkpoint.filepath, os.path.join(model.model_dir, 'best.hdf5'))
        self.assertEqual(csv_logger.filename, os.path.join(model.model_dir, 'logger.csv'))
        self.assertEqual(early_stopping.patience, 12)

        # level save 'LOW'
        model.level_save = 'LOW'
        callbacks = model._get_callbacks(patience=14)
        callbacks_types = [type(_) for _ in callbacks]
        self.assertTrue(tf.keras.callbacks.EarlyStopping in callbacks_types)
        self.assertFalse(ModelCheckpointAll in callbacks_types)
        self.assertTrue(tf.keras.callbacks.CSVLogger in callbacks_types)
        self.assertTrue(tf.keras.callbacks.TerminateOnNaN in callbacks_types)
        csv_logger = callbacks[callbacks_types.index(tf.keras.callbacks.CSVLogger)]
        early_stopping = callbacks[callbacks_types.index(tf.keras.callbacks.EarlyStopping)]
        self.assertEqual(csv_logger.filename, os.path.join(model.model_dir, 'logger.csv'))
        self.assertEqual(early_stopping.patience, 14)

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test14_model_keras_faster_rcnn_get_learning_rate_scheduler(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._get_learning_rate_scheduler'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        self.assertEqual(model._get_learning_rate_scheduler(), None)

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test15_model_keras_faster_rcnn_predict_object_detector(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._predict_object_detector'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        df_data = pd.DataFrame({'file_path': path_list, 'bboxes': bboxes_list}).copy()

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10)
        model.fit(df_data, df_valid=df_data)
        preds = model.predict(df_data)
        self.assertEqual(len(preds), df_data.shape[0])
        # We can't test lots of things ...
        remove_dir(model_dir)

        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
            model.predict(df_data)

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test16_model_keras_faster_rcnn_rcnn_plot_metrics_and_loss(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn_rcnn.ModelDetectronFasterRcnnObjectDetector._plot_metrics_and_loss'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        class FitHistory():
            def __init__(self, history) -> None:
                self.history = history
        fit_history = FitHistory({
            'loss': [1, 2, 3], 'val_loss': [1, 2, 3],
            'rpn_class_loss': [1, 2, 3], 'val_rpn_class_loss': [1, 2, 3],
            'rpn_regr_loss': [1, 2, 3], 'val_rpn_regr_loss': [1, 2, 3],
            'rpn_class_accuracy': [1, 2, 3], 'val_rpn_class_accuracy': [1, 2, 3],
            'dense_class_loss': [1, 2, 3], 'val_dense_class_loss': [1, 2, 3],
            'dense_regr_loss': [1, 2, 3], 'val_dense_regr_loss': [1, 2, 3],
            'dense_class_accuracy': [1, 2, 3], 'val_dense_class_accuracy': [1, 2, 3],
        })
        fit_history_small = FitHistory({
            'loss': [1, 2, 3], 'val_loss': [1, 2, 3],
            'rpn_regr_loss': [1, 2, 3], 'val_rpn_regr_loss': [1, 2, 3],
        })
        #
        targets_name =  {f'{model_type}_{trainable}': [
                            f'loss_{model_type}_trainable_{trainable}',
                            f'loss_class_{model_type}_trainable_{trainable}',
                            f'loss_regr_{model_type}_trainable_{trainable}',
                            f'accuracy_class_{model_type}_trainable_{trainable}',
                        ] for model_type in ['rpn', 'classifier'] for trainable in [False, True]}

        # Set a model
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        plots_path = os.path.join(model.model_dir, 'plots')

        # Nominal case
        model._plot_metrics_and_loss(fit_history, model_type='rpn', trainable=True)
        list_files = os.listdir(plots_path)
        self.assertEqual(len(list_files), 4)
        for filename in targets_name['rpn_True']:
            self.assertTrue(os.path.exists(os.path.join(plots_path, f"{filename}.jpeg")))
        remove_dir(plots_path)
        #
        model._plot_metrics_and_loss(fit_history, model_type='classifier', trainable=True)
        list_files = os.listdir(plots_path)
        self.assertEqual(len(list_files), 4)
        for filename in targets_name['classifier_True']:
            self.assertTrue(os.path.exists(os.path.join(plots_path, f"{filename}.jpeg")))
        remove_dir(plots_path)
        #
        model._plot_metrics_and_loss(fit_history, model_type='rpn', trainable=False)
        list_files = os.listdir(plots_path)
        self.assertEqual(len(list_files), 4)
        for filename in targets_name['rpn_False']:
            self.assertTrue(os.path.exists(os.path.join(plots_path, f"{filename}.jpeg")))
        remove_dir(plots_path)
        #
        model._plot_metrics_and_loss(fit_history, model_type='classifier', trainable=False)
        list_files = os.listdir(plots_path)
        self.assertEqual(len(list_files), 4)
        for filename in targets_name['classifier_False']:
            self.assertTrue(os.path.exists(os.path.join(plots_path, f"{filename}.jpeg")))
        remove_dir(plots_path)

        # Less metrics
        model._plot_metrics_and_loss(fit_history_small, model_type='rpn', trainable=True)
        list_files = os.listdir(plots_path)
        self.assertEqual(len(list_files), 2)
        remove_dir(plots_path)

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test17_model_keras_faster_rcnn_save(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn_rcnn.ModelDetectronFasterRcnnObjectDetector.save'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        df_data = pd.DataFrame({'file_path': path_list, 'bboxes': bboxes_list}).copy()

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10)
        model.fit(df_data)
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"preprocess_input.pkl")))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertTrue('maintainers' in configs.keys())
        self.assertTrue('date' in configs.keys())
        self.assertTrue('package_version' in configs.keys())
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertTrue('model_name' in configs.keys())
        self.assertTrue('model_dir' in configs.keys())
        self.assertTrue('model_type' in configs.keys())
        self.assertTrue('trained' in configs.keys())
        self.assertTrue('nb_fit' in configs.keys())
        self.assertTrue('level_save' in configs.keys())
        self.assertTrue('librairie' in configs.keys())
        self.assertEqual(configs['librairie'], 'keras')
        self.assertTrue('batch_size' in configs.keys())
        self.assertTrue('epochs' in configs.keys())
        self.assertTrue('validation_split' in configs.keys())
        self.assertTrue('patience' in configs.keys())
        self.assertTrue('width' in configs.keys())
        self.assertTrue('height' in configs.keys())
        self.assertTrue('depth' in configs.keys())
        self.assertTrue('color_mode' in configs.keys())
        self.assertTrue('in_memory' in configs.keys())
        self.assertTrue('data_augmentation_params' in configs.keys())
        self.assertTrue('nb_train_generator_images_to_save' in configs.keys())
        self.assertTrue('keras_params' in configs.keys())
        self.assertTrue('_get_model' in configs.keys())
        self.assertTrue('_get_learning_rate_scheduler' in configs.keys())
        self.assertTrue('custom_objects' in configs.keys())
        self.assertTrue('list_classes' in configs.keys())
        self.assertTrue('dict_classes' in configs.keys())
        self.assertTrue('vgg_filename' in configs.keys())
        self.assertTrue('shared_model_subsampling' in configs.keys())
        self.assertTrue('anchor_box_sizes' in configs.keys())
        self.assertTrue('anchor_box_ratios' in configs.keys())
        self.assertTrue('nb_anchors' in configs.keys())
        self.assertTrue('list_anchors' in configs.keys())
        self.assertTrue('img_min_side_size' in configs.keys())
        self.assertTrue('pool_resize_classifier' in configs.keys())
        self.assertTrue('rpn_regr_scaling' in configs.keys())
        self.assertTrue('classifier_regr_scaling' in configs.keys())
        self.assertTrue('rpn_min_overlap' in configs.keys())
        self.assertTrue('rpn_max_overlap' in configs.keys())
        self.assertTrue('rpn_restrict_num_regions' in configs.keys())
        self.assertTrue('nb_rois_classifier' in configs.keys())
        self.assertTrue('roi_nms_overlap_threshold' in configs.keys())
        self.assertTrue('nms_max_boxes' in configs.keys())
        self.assertTrue('classifier_min_overlap' in configs.keys())
        self.assertTrue('classifier_max_overlap' in configs.keys())
        self.assertTrue('pred_bbox_proba_threshold' in configs.keys())
        self.assertTrue('pred_nms_overlap_threshold' in configs.keys())
        self.assertTrue('batch_size_rpn_trainable_true' in configs.keys())
        self.assertTrue('batch_size_classifier_trainable_true' in configs.keys())
        self.assertTrue('batch_size_rpn_trainable_false' in configs.keys())
        self.assertTrue('batch_size_classifier_trainable_false' in configs.keys())
        self.assertTrue('epochs_rpn_trainable_true' in configs.keys())
        self.assertTrue('epochs_classifier_trainable_true' in configs.keys())
        self.assertTrue('epochs_rpn_trainable_false' in configs.keys())
        self.assertTrue('epochs_classifier_trainable_false' in configs.keys())
        self.assertTrue('patience_rpn_trainable_true' in configs.keys())
        self.assertTrue('patience_classifier_trainable_true' in configs.keys())
        self.assertTrue('patience_rpn_trainable_false' in configs.keys())
        self.assertTrue('patience_classifier_trainable_false' in configs.keys())
        self.assertTrue('_add_rpn_layers' in configs.keys())
        self.assertTrue('_add_classifier_layers' in configs.keys())
        self.assertTrue('_compile_model_rpn' in configs.keys())
        self.assertTrue('_compile_model_classifier' in configs.keys())
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test18_model_keras_faster_rcnn_reload_model(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn_rcnn.ModelDetectronFasterRcnnObjectDetector.reload_model'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        df_data = pd.DataFrame({'file_path': path_list, 'bboxes': bboxes_list}).copy()


        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10)
        model.fit(df_data)
        model.save()
        # Reload keras
        hdf5_path = os.path.join(model.model_dir, 'best.hdf5')
        reloaded_model = model.reload_model(hdf5_path)
        # We can't really test the predictions because chances are high that no bboxes are predicted

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test19_model_keras_faster_rcnn_reload_from_standalone(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn_rcnn.ModelDetectronFasterRcnnObjectDetector.reload_from_standalone'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        model_dir_2 = os.path.join(os.getcwd(), 'model_test_123456789_2')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        df_data = pd.DataFrame({'file_path': path_list, 'bboxes': bboxes_list}).copy()


        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10)
        fit_params = model.fit(df_data)
        model.save()
        # Reload
        conf_path = os.path.join(model.model_dir, "configurations.json")
        hdf5_path = os.path.join(model.model_dir, "best.hdf5")
        preprocess_input_path = os.path.join(model.model_dir, "preprocess_input.pkl")
        new_model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir_2)
        new_model.reload_from_standalone(configuration_path=conf_path, hdf5_path=hdf5_path,
                                         preprocess_input_path=preprocess_input_path)
        # Tests
        self.assertEqual(model.model_name, new_model.model_name)
        self.assertEqual(model.model_type, new_model.model_type)
        self.assertEqual(model.trained, new_model.trained)
        self.assertEqual(model.nb_fit, new_model.nb_fit)
        self.assertEqual(model.level_save, new_model.level_save)
        self.assertEqual(model.list_classes, new_model.list_classes)
        self.assertEqual(model.dict_classes, new_model.dict_classes)
        self.assertEqual(model.batch_size, new_model.batch_size)
        self.assertEqual(model.epochs, new_model.epochs)
        self.assertEqual(model.validation_split, new_model.validation_split)
        self.assertEqual(model.patience, new_model.patience)
        self.assertEqual(model.width, new_model.width)
        self.assertEqual(model.height, new_model.height)
        self.assertEqual(model.depth, new_model.depth)
        self.assertEqual(model.color_mode, new_model.color_mode)
        self.assertEqual(model.in_memory, new_model.in_memory)
        self.assertEqual(model.data_augmentation_params, new_model.data_augmentation_params)
        self.assertEqual(model.nb_train_generator_images_to_save, new_model.nb_train_generator_images_to_save)
        self.assertEqual(model.vgg_filename, new_model.vgg_filename)
        self.assertEqual(model.vgg_path, new_model.vgg_path)
        self.assertEqual(model.shared_model_subsampling, new_model.shared_model_subsampling)
        self.assertEqual(model.anchor_box_sizes, new_model.anchor_box_sizes)
        self.assertEqual(model.anchor_box_ratios, new_model.anchor_box_ratios)
        self.assertEqual(model.nb_anchors, new_model.nb_anchors)
        self.assertEqual(model.list_anchors, new_model.list_anchors)
        self.assertEqual(model.img_min_side_size, new_model.img_min_side_size)
        self.assertEqual(model.pool_resize_classifier, new_model.pool_resize_classifier)
        self.assertEqual(model.rpn_regr_scaling, new_model.rpn_regr_scaling)
        self.assertEqual(model.classifier_regr_scaling, new_model.classifier_regr_scaling)
        self.assertEqual(model.rpn_min_overlap, new_model.rpn_min_overlap)
        self.assertEqual(model.rpn_max_overlap, new_model.rpn_max_overlap)
        self.assertEqual(model.rpn_restrict_num_regions, new_model.rpn_restrict_num_regions)
        self.assertEqual(model.nb_rois_classifier, new_model.nb_rois_classifier)
        self.assertEqual(model.roi_nms_overlap_threshold, new_model.roi_nms_overlap_threshold)
        self.assertEqual(model.nms_max_boxes, new_model.nms_max_boxes)
        self.assertEqual(model.classifier_min_overlap, new_model.classifier_min_overlap)
        self.assertEqual(model.classifier_max_overlap, new_model.classifier_max_overlap)
        self.assertEqual(model.pred_bbox_proba_threshold, new_model.pred_bbox_proba_threshold)
        self.assertEqual(model.pred_nms_overlap_threshold, new_model.pred_nms_overlap_threshold)
        self.assertEqual(model.batch_size_rpn_trainable_true, new_model.batch_size_rpn_trainable_true)
        self.assertEqual(model.batch_size_classifier_trainable_true, new_model.batch_size_classifier_trainable_true)
        self.assertEqual(model.batch_size_classifier_trainable_true, new_model.batch_size_classifier_trainable_true)
        self.assertEqual(model.batch_size_rpn_trainable_false, new_model.batch_size_rpn_trainable_false)
        self.assertEqual(model.batch_size_rpn_trainable_false, new_model.batch_size_rpn_trainable_false)
        self.assertEqual(model.batch_size_classifier_trainable_false, new_model.batch_size_classifier_trainable_false)
        self.assertEqual(model.batch_size_classifier_trainable_false, new_model.batch_size_classifier_trainable_false)
        self.assertEqual(model.epochs_rpn_trainable_true, new_model.epochs_rpn_trainable_true)
        self.assertEqual(model.epochs_classifier_trainable_true, new_model.epochs_classifier_trainable_true)
        self.assertEqual(model.epochs_classifier_trainable_true, new_model.epochs_classifier_trainable_true)
        self.assertEqual(model.epochs_rpn_trainable_false, new_model.epochs_rpn_trainable_false)
        self.assertEqual(model.epochs_classifier_trainable_false, new_model.epochs_classifier_trainable_false)
        self.assertEqual(model.epochs_classifier_trainable_false, new_model.epochs_classifier_trainable_false)
        self.assertEqual(model.patience_rpn_trainable_true, new_model.patience_rpn_trainable_true)
        self.assertEqual(model.patience_classifier_trainable_true, new_model.patience_classifier_trainable_true)
        self.assertEqual(model.patience_classifier_trainable_true, new_model.patience_classifier_trainable_true)
        self.assertEqual(model.patience_rpn_trainable_false, new_model.patience_rpn_trainable_false)
        self.assertEqual(model.patience_classifier_trainable_false, new_model.patience_classifier_trainable_false)
        self.assertEqual(model.patience_classifier_trainable_false, new_model.patience_classifier_trainable_false)
        self.assertEqual(model.keras_params, new_model.keras_params)
        # self.assertEqual(model.custom_objects, new_model.custom_objects)  # Do not work because functions are wrapped ...
        # We can't really test the pipeline so we test predictions
        self.assertEqual(model.predict(df_data), new_model.predict(df_data))  # Useful ?
        remove_dir(model_dir)
        remove_dir(new_model.model_dir)


        ############################################
        # Errors
        ############################################

        # with self.assertRaises(FileNotFoundError):
        #     new_model = ModelCnnClassifier(model_dir=model_dir_2)
        #     new_model.reload_from_standalone(configuration_path='toto.json', hdf5_path=hdf5_path,
        #                                      preprocess_input_path=preprocess_input_path)
        # with self.assertRaises(FileNotFoundError):
        #     new_model = ModelCnnClassifier(model_dir=model_dir_2)
        #     new_model.reload_from_standalone(configuration_path=conf_path, hdf5_path='toto.pkl',
        #                                      preprocess_input_path=preprocess_input_path)
        # with self.assertRaises(FileNotFoundError):
        #     new_model = ModelCnnClassifier(model_dir=model_dir_2)
        #     new_model.reload_from_standalone(configuration_path=conf_path, hdf5_path=hdf5_path,
        #                                      preprocess_input_path='toto.pkl')

        # Clean
        remove_dir(model_dir)


class CustomGeneratorRpnTests(unittest.TestCase):
    '''Main class to test model_keras_faster_rcnn.CustomGeneratorRpn'''

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    @patch('{{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector', side_effect=MockModelKerasFasterRcnnObjectDetector)
    def test01_custom_generator_rpn_init(self, mock_model):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.CustomGeneratorRpn.__init__'''

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        img_data_list = []
        for filepath, bboxes in zip(path_list, bboxes_list):
            img_data_list.append({'file_path': filepath, 'bboxes': bboxes})
        fake_model = mock_model()

        # Init., test all parameters
        generator = CustomGeneratorRpn(img_data_list, batch_size=5, shuffle=True, seed=42, model=fake_model,
                                       horizontal_flip=False, vertical_flip=False, rot_90=False, data_type='train',
                                       with_img_data=False)
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, True)
        self.assertEqual(generator.seed, 42)
        self.assertEqual(generator.img_data_list, img_data_list)
        self.assertEqual(generator.model, fake_model)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.with_img_data, False)
        self.assertEqual(generator.is_test, False)
        # Other params
        generator = CustomGeneratorRpn(img_data_list, batch_size=10, shuffle=False, seed=64, model=fake_model,
                                       horizontal_flip=True, vertical_flip=True, rot_90=True, data_type='test',
                                       with_img_data=True)
        self.assertEqual(generator.batch_size, 1)  # test data, force batch size 1
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, 64)
        self.assertEqual(generator.img_data_list, img_data_list)
        self.assertEqual(generator.model, fake_model)
        self.assertEqual(generator.horizontal_flip, True)
        self.assertEqual(generator.vertical_flip, True)
        self.assertEqual(generator.rot_90, True)
        self.assertEqual(generator.with_img_data, True)
        self.assertEqual(generator.is_test, True)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    @patch('{{package_name}}.models_training.object_detectors.utils_object_detectors.get_rpn_targets', side_effect=mock_get_rpn_targets)
    @patch('{{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector', side_effect=MockModelKerasFasterRcnnObjectDetector)
    def test02_custom_generator_rpn_get_batches_of_transformed_samples(self, mock_model, mock_rpn_targets):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.CustomGeneratorRpn._get_batches_of_transformed_samples'''

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        img_data_list = []
        for filepath, bboxes in zip(path_list, bboxes_list):
            img_data_list.append({'file_path': filepath, 'bboxes': bboxes})
        fake_model = mock_model()

        # Nominal case
        generator = CustomGeneratorRpn(img_data_list, batch_size=2, shuffle=True, seed=42, model=fake_model,
                                       data_type='train', with_img_data=False)
        X, Y = generator._get_batches_of_transformed_samples(index_array=np.array([3, 1]))
        self.assertEqual(list(X.keys()), ['input_img'])
        self.assertEqual(X['input_img'].shape[0], 2)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(sorted(list(Y.keys())), sorted(['rpn_class', 'rpn_regr']))
        # with_img_data = True
        generator = CustomGeneratorRpn(img_data_list, batch_size=2, shuffle=True, seed=42, model=fake_model,
                                       data_type='train', with_img_data=True)
        X, Y, batch_data = generator._get_batches_of_transformed_samples(index_array=np.array([3, 1]))
        self.assertEqual(list(X.keys()), ['input_img'])
        self.assertEqual(X['input_img'].shape[0], 2)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(sorted(list(Y.keys())), sorted(['rpn_class', 'rpn_regr']))
        self.assertEqual(len(batch_data), 2)
        self.assertTrue('batch_width' in batch_data[0].keys())
        self.assertTrue('batch_height' in batch_data[0].keys())
        # data_type == 'test'
        generator = CustomGeneratorRpn(img_data_list, batch_size=2, shuffle=True, seed=42, model=fake_model,
                                       data_type='test', with_img_data=False)
        X = generator._get_batches_of_transformed_samples(index_array=np.array([2]))
        self.assertEqual(list(X.keys()), ['input_img'])
        self.assertEqual(X['input_img'].shape[0], 1)
        self.assertEqual(X['input_img'].shape[-1], 3)
        # data_type == 'test' & with_img_data == True
        generator = CustomGeneratorRpn(img_data_list, batch_size=2, shuffle=True, seed=42, model=fake_model,
                                       data_type='test', with_img_data=True)
        X, batch_data = generator._get_batches_of_transformed_samples(index_array=np.array([2]))
        self.assertEqual(list(X.keys()), ['input_img'])
        self.assertEqual(X['input_img'].shape[0], 1)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(len(batch_data), 1)
        self.assertTrue('batch_width' in batch_data[0].keys())
        self.assertTrue('batch_height' in batch_data[0].keys())


class CustomGeneratorClassifierTests(unittest.TestCase):
    '''Main class to test model_keras_faster_rcnn.CustomGeneratorClassifier'''

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    @patch('{{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.clone_model', side_effect=mock_clone_model_fn)
    @patch('{{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector', side_effect=MockModelKerasFasterRcnnObjectDetector)
    def test01_custom_generator_classifier_init(self, mock_model, mock_clone_model):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.CustomGeneratorClassifier.__init__'''

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        img_data_list = []
        for filepath, bboxes in zip(path_list, bboxes_list):
            img_data_list.append({'file_path': filepath, 'bboxes': bboxes})
        fake_model = mock_model()

        # Init., test all parameters
        generator = CustomGeneratorClassifier(img_data_list, batch_size=5, shuffle=True, seed=42, model=fake_model, shared_model_trainable=False,
                                              horizontal_flip=False, vertical_flip=False, rot_90=False, data_type='train', with_img_data=False)
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, True)
        self.assertEqual(generator.seed, 42)
        self.assertEqual(generator.img_data_list, img_data_list)
        self.assertEqual(generator.model, fake_model)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.with_img_data, False)
        self.assertEqual(generator.is_test, False)
        self.assertTrue(generator.rpn_clone is None)
        # Other params
        generator = CustomGeneratorClassifier(img_data_list, batch_size=10, shuffle=False, seed=64, model=fake_model, shared_model_trainable=True,
                                              horizontal_flip=True, vertical_flip=True, rot_90=True, data_type='test', with_img_data=True)
        self.assertEqual(generator.batch_size, 1)  # test data, force batch size 1
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, 64)
        self.assertEqual(generator.img_data_list, img_data_list)
        self.assertEqual(generator.model, fake_model)
        self.assertEqual(generator.horizontal_flip, True)
        self.assertEqual(generator.vertical_flip, True)
        self.assertEqual(generator.rot_90, True)
        self.assertEqual(generator.with_img_data, True)
        self.assertEqual(generator.is_test, True)
        self.assertFalse(generator.rpn_clone is None)

    # We do not mock anything, too complex
    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test02_custom_generator_classifier_get_batches_of_transformed_samples(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.CustomGeneratorClassifier._get_batches_of_transformed_samples'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars & model
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        df_data = pd.DataFrame({'file_path': path_list, 'bboxes': bboxes_list}).copy()
        img_data_list = []
        for filepath, bboxes in zip(path_list, bboxes_list):
            img_data_list.append({'file_path': filepath, 'bboxes': bboxes})
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10, nb_rois_classifier=3)
        model.fit(df_data)

        # Nominal case
        generator = CustomGeneratorClassifier(img_data_list, batch_size=2, shuffle=True, seed=42, model=model,
                                              shared_model_trainable=False, data_type='train', with_img_data=False)
        X, Y = generator._get_batches_of_transformed_samples(index_array=np.array([3, 1]))
        self.assertEqual(sorted(list(X.keys())), sorted(['input_img', 'input_rois']))
        self.assertEqual(X['input_img'].shape[0], 2)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(X['input_rois'].shape[0], 2)
        self.assertEqual(X['input_rois'].shape[1], model.nb_rois_classifier)
        self.assertEqual(X['input_rois'].shape[2], 4)
        self.assertEqual(sorted(list(Y.keys())), sorted(['dense_class', 'dense_regr']))
        self.assertEqual(Y['dense_class'].shape[0], 2)
        self.assertEqual(Y['dense_class'].shape[1], model.nb_rois_classifier)
        self.assertEqual(Y['dense_regr'].shape[0], 2)
        self.assertEqual(Y['dense_regr'].shape[1], model.nb_rois_classifier)
        # with_img_data = True & shared_model_trainable à True
        generator = CustomGeneratorClassifier(img_data_list, batch_size=2, shuffle=True, seed=42, model=model,
                                              shared_model_trainable=True, data_type='train', with_img_data=True)
        X, Y, batch_data = generator._get_batches_of_transformed_samples(index_array=np.array([3, 1]))
        self.assertEqual(sorted(list(X.keys())), sorted(['input_img', 'input_rois']))
        self.assertEqual(X['input_img'].shape[0], 2)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(X['input_rois'].shape[0], 2)
        self.assertEqual(X['input_rois'].shape[1], model.nb_rois_classifier)
        self.assertEqual(X['input_rois'].shape[2], 4)
        self.assertEqual(sorted(list(Y.keys())), sorted(['dense_class', 'dense_regr']))
        self.assertEqual(Y['dense_class'].shape[0], 2)
        self.assertEqual(Y['dense_class'].shape[1], model.nb_rois_classifier)
        self.assertEqual(Y['dense_regr'].shape[0], 2)
        self.assertEqual(Y['dense_regr'].shape[1], model.nb_rois_classifier)
        self.assertEqual(len(batch_data), 2)
        self.assertTrue('batch_width' in batch_data[0].keys())
        self.assertTrue('batch_height' in batch_data[0].keys())
        # data_type == 'test'
        generator = CustomGeneratorClassifier(img_data_list, batch_size=2, shuffle=True, seed=42, model=model,
                                              shared_model_trainable=True, data_type='test', with_img_data=False)
        X = generator._get_batches_of_transformed_samples(index_array=np.array([2]))
        self.assertEqual(sorted(list(X.keys())), sorted(['input_img', 'input_rois']))
        self.assertEqual(X['input_img'].shape[0], 1)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(X['input_rois'].shape[0], 1)
        # self.assertEqual(X['input_rois'].shape[1], ???) -> We do not know the number of ROIs returned for the test
        self.assertEqual(X['input_rois'].shape[2], 4)
        # data_type == 'test' & with_img_data == True
        generator = CustomGeneratorClassifier(img_data_list, batch_size=2, shuffle=True, seed=42, model=model,
                                              shared_model_trainable=True, data_type='test', with_img_data=True)
        X, batch_data = generator._get_batches_of_transformed_samples(index_array=np.array([2]))
        self.assertEqual(sorted(list(X.keys())), sorted(['input_img', 'input_rois']))
        self.assertEqual(X['input_img'].shape[0], 1)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(X['input_rois'].shape[0], 1)
        # self.assertEqual(X['input_rois'].shape[1], ???) -> We do not know the number of ROIs returned for the test
        self.assertEqual(X['input_rois'].shape[2], 4)
        self.assertEqual(len(batch_data), 1)
        self.assertTrue('batch_width' in batch_data[0].keys())
        self.assertTrue('batch_height' in batch_data[0].keys())


# We do not test ModelCheckpointAll


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3

## Split a dataset into training / validation / test sets
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Ex: python 0_split_train_valid_test.py -f dataset.csv --perc_train 0.6 --perc_valid 0.2 --perc_test 0.2


import os
import random
import logging
import argparse
import numpy as np
import pandas as pd
from pathlib import Path
from typing import Union, Tuple

from {{package_name}} import utils
from {{package_name}}.models_training import utils_models


# Get logger
logger = logging.getLogger('{{package_name}}.0_split_train_valid_test')


def main(filename: str, split_type: str, perc_train: float, perc_valid: float, perc_test: float,
         y_col: Union[str, int, None] = None, overwrite_dataset: bool = False,
         seed: Union[int, None] = None, sep: str = '{{default_sep}}', encoding: str = '{{default_encoding}}') -> None:
    '''Splits a dataset into training / validation / test sets

    Args:
        filename (str): Name of the dataset to be splitted (actually a path relative to {{package_name}}-data)
        split_type (str): Split type (random, stratified)
        perc_train (float): Train dataset fraction
        perc_valid (float): Validation dataset fraction
        perc_test (float): Test dataset fraction
    Kwargs:
        y_col (str | int): Column to be used with stratified split
        overwrite_dataset (bool): Whether to allow overwriting datasets
        seed (int): Random seed
        sep (str): Separator to use with the .csv files
        encoding (str): Encoding to use with the .csv files
    Raises:
        ValueError: If split_type value is not a valid option (['random', 'stratified'])
        ValueError: If stratified split, but y_col is None
        ValueError: If perc_train not in in ]0, 1[
        ValueError: If perc_valid not in in [0, 1[
        ValueError: If perc_test not in in [0, 1[
        ValueError: If abs(perc_train + perc_valid + perc_test - 1) > 0.0001
        ValueError: If perc_valid & perc_test are both null
        FileNotFoundError: If the file does not exist in {{package_name}}-data
    '''
    logger.info(f"Splits {filename} into training / validation / test sets ...")

    # Manage errors
    if split_type not in ['random', 'stratified']:
        raise ValueError(f"split_type value ({split_type}) is not a valid option (['random', 'stratified']).")
    if split_type == 'stratified' and y_col is None:
        raise ValueError("y_col must be set with 'stratified' split option")
    if not 0 < perc_train < 1:
        raise ValueError("perc_train must be positive in ]0, 1[")
    if not 0 <= perc_valid < 1:
        raise ValueError("perc_valid must be positive in [0, 1[")
    if not 0 <= perc_test < 1:
        raise ValueError("perc_test must be positive in [0, 1[")
    if abs(perc_train + perc_valid + perc_test - 1) > 0.0001:
        raise ValueError(f"The sum of perc_train, perc_valid and perc_test should be equal to 1, not {perc_train + perc_valid + perc_test}")
    if perc_valid == 0. and perc_test == 0.:
        raise ValueError("perc_valid & perc_test can't both be null.")

    # Set seed
    if seed is not None:
        random.seed(seed)
        np.random.seed(seed)

    # Manage paths
    data_path = utils.get_data_path()
    file_path = os.path.join(data_path, filename)
    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"The file {file_path} does not exist")
    basename = Path(filename).stem

    # Manage save paths
    train_path =  os.path.join(data_path, f"{basename}_train.csv")
    valid_path =  os.path.join(data_path, f"{basename}_valid.csv")
    test_path =  os.path.join(data_path, f"{basename}_test.csv")
    if os.path.exists(train_path) and not overwrite_dataset:
        logger.warning(f"{train_path} already exists. Use --overwrite to overwrite it.")
        return
    if os.path.exists(valid_path) and not overwrite_dataset and perc_valid != 0.:
        logger.warning(f"{valid_path} already exists. Use --overwrite to overwrite it.")
        return
    if os.path.exists(test_path) and not overwrite_dataset and perc_test != 0.:
        logger.warning(f"{test_path} already exists. Use --overwrite to overwrite it.")
        return

    # Get dataframe
    # We load everything as string & fillna (avoid some errors)
    # TODO: check if this is correct
    df, first_line = utils.read_csv(file_path, sep=sep, encoding=encoding, dtype=str)
    df = df.fillna('')
    nb_lines = df.shape[0]

    # Percentages
    logger.info(f'Train percentage : {perc_train * 100}%')
    logger.info(f'Validation percentage : {perc_valid * 100}%')
    logger.info(f'Test percentage : {perc_test * 100}%')

    # Split
    if split_type == 'random':
        df_train, df_valid, df_test = split_random(df, perc_train, perc_valid, perc_test, seed=seed)
    else:  # stratified
        df_train, df_valid, df_test = split_stratified(df, y_col, perc_train, perc_valid, perc_test, seed=seed)

    # Display info
    logger.info(f"Number of lines in the original dataset : {nb_lines}")
    logger.info(f"Number of lines in the train dataset : {df_train.shape[0]} ({df_train.shape[0] / nb_lines * 100} %)")
    if df_valid is not None:
        logger.info(f"Number of lines in the validation dataset : {df_valid.shape[0]} ({df_valid.shape[0] / nb_lines * 100} %)")
    else:
        logger.info("No validation dataset generated (perc. at 0.0% )")
    if df_test is not None:
        logger.info(f"Number of lines in the test dataset : {df_test.shape[0]} ({df_test.shape[0] / nb_lines * 100} %)")
    else:
        logger.info("No test dataset generated (perc. at 0.0% )")

    # Save
    utils.to_csv(df_train, train_path, first_line=first_line, sep='{{default_sep}}', encoding='{{default_encoding}}')
    if df_valid is not None:
        utils.to_csv(df_valid, valid_path, first_line=first_line, sep='{{default_sep}}', encoding='{{default_encoding}}')
    if df_test is not None:
        utils.to_csv(df_test, test_path, first_line=first_line, sep='{{default_sep}}', encoding='{{default_encoding}}')


def split_random(df: pd.DataFrame, perc_train: float, perc_valid: float, perc_test: float, seed: Union[int, None] = None)\
        -> Tuple[pd.DataFrame, Union[pd.DataFrame, None], Union[pd.DataFrame, None]]:
    '''Splits a dataset into training / validation / test sets - `Random` strategy

    Args:
        df (pd.DataFrame): Dataset to be splitted
        perc_train (float): Train dataset fraction
        perc_valid (float): Validation dataset fraction
        perc_test (float): Test dataset fraction
    Kwargs:
        seed (int): Seed to use
    Returns:
        pd.DataFrame: Train dataset
        pd.DataFrame: Validation dataset
        pd.DataFrame: Test dataset
    '''
    # Select train
    df_train = df.sample(frac=perc_train, random_state=seed)
    idx_train = df_train.index
    df = df[~df.index.isin(idx_train)]  # Remove train from main data
    # Select valid
    if perc_valid > 0.0:
        if perc_test == 0.0:
            df_valid = df  # no test, remaining data is validation
        else:
            # Update perc_valid to match remaining data
            perc_valid = perc_valid / (perc_test + perc_valid)
            df_valid = df.sample(frac=perc_valid, random_state=seed)
            idx_valid = df_valid.index
            df = df[~df.index.isin(idx_valid)]  # Remove valid from main data
    else:
        df_valid = None
    # Select test - remaining data
    df_test = df if perc_test > 0.0 else None
    # Return
    return df_train, df_valid, df_test


def split_stratified(df: pd.DataFrame, y_col: Union[str, int], perc_train: float, perc_valid: float, perc_test: float,
                     seed: Union[int, None] = None) -> Tuple[pd.DataFrame, Union[pd.DataFrame, None], Union[pd.DataFrame, None]]:
    '''Splits a dataset into training / validation / test sets - `Stratified` strategy

    Args:
        df (pd.DataFrame): Dataset to be splitted
        y_col (str | int): Column to be used with stratified split
        perc_train (float): Train dataset fraction
        perc_valid (float): Validation dataset fraction
        perc_test (float): Test dataset fraction
    Kwargs:
        seed (int): Seed to use
    Returns:
        pd.DataFrame: Train dataset
        pd.DataFrame: Validation dataset
        pd.DataFrame: Test dataset
    '''
    # Select train
    df_train, df_valid_test = utils_models.stratified_split(df, y_col, test_size=perc_valid + perc_test, seed=seed)
    # Nominal case
    if perc_valid > 0 and perc_test > 0:
        # Update perc_test to match remaining data
        perc_test = perc_test / (perc_test + perc_valid)
        df_valid, df_test = utils_models.stratified_split(df_valid_test, y_col, test_size=perc_test, seed=seed)
        # Some indexes might be "lost" (small_classes), we retrieve them
        df_lost = df[~df.index.isin(df_train.index.append(df_test.index).append(df_valid.index))]
    # No validation set
    elif perc_valid == 0.0:
        df_valid = None
        df_test = df_valid_test
        # Some indexes might be "lost" (small_classes), we retrieve them
        df_lost = df[~df.index.isin(df_train.index.append(df_test.index))]
    # No test set
    else:
        df_valid = df_valid_test
        df_test = None
        # Some indexes might be "lost" (small_classes), we retrieve them
        df_lost = df[~df.index.isin(df_train.index.append(df_valid.index))]
    # We finally append "lost" indexes to the training set
    df_train = pd.concat([df_train, df_lost], sort=False)
    # Return
    return df_train, df_valid, df_test


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--filename', default='dataset.csv', help="Name of the dataset to be splitted (actually a path relative to {{package_name}}-data)")
    parser.add_argument('--split_type', default='random', help="Split type (random, stratified)")
    parser.add_argument('--perc_train', default=0.6, type=float, help="Train dataset fraction")
    parser.add_argument('--perc_valid', default=0.2, type=float, help="Validation dataset fraction")
    parser.add_argument('--perc_test', default=0.2, type=float, help="Test dataset fraction")
    parser.add_argument('--y_col', default=None, help="Column to be used with stratified split")
    parser.add_argument('--overwrite', dest='overwrite_dataset', action='store_true', help="Whether to allow overwriting datasets")
    parser.add_argument('--seed', default=None, type=int, help="Random seed")
    parser.add_argument('--sep', default='{{default_sep}}', help="Separator to use with the .csv files")
    parser.add_argument('--encoding', default="{{default_encoding}}", help="Encoding to use with the .csv files")
    parser.set_defaults(overwrite_dataset=False)
    args = parser.parse_args()
    main(filename=args.filename, split_type=args.split_type, perc_train=args.perc_train,
         perc_valid=args.perc_valid, perc_test=args.perc_test, y_col=args.y_col,
         overwrite_dataset=args.overwrite_dataset, seed=args.seed, sep=args.sep,
         encoding=args.encoding)



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""This module contains a ModelGabarit class you can use for your gabarit generated
projects

ModelGabarit overwrite some methods of the base Model class :
    - download_model method to download a model from a JFrog Artifactory repository ;
    {%- if gabarit_package_spec %}
    - _load_model method to use the {{gabarit_import_name}}.models_training.utils_models.load_model
    function from a typical gabarit project ;
    - predict method to use the the {{gabarit_import_name}}.models_training.utils_models.predict
    function from a typical gabarit project.
    {%- else %}
    - _load_model method to use the gabarit_package.models_training.utils_models.load_model
    function from a typical gabarit project ;
    - predict method to use the the gabarit_package.models_training.utils_models.predict
    function from a typical gabarit project.
    {%- endif %}

"""


import logging
import shutil
import tempfile
import pandas as pd
from pathlib import Path
from typing import Any, Union
from pydantic import BaseSettings

from .model_base import Model

{%- if gabarit_package_spec %}
try:
    from {{gabarit_import_name}} import utils as utils_gabarit
    from {{gabarit_import_name}}.models_training import utils_models
    from {{gabarit_import_name}}.monitoring.model_explainer import Explainer
except ImportError:
    raise ImportError("Package '{{gabarit_package_name}}' not found. Please install it.")
{%- else %}
try:
    from gabarit_package import utils as utils_gabarit
    from gabarit_package.models_training import utils_models
    from gabarit_package.monitoring.model_explainer import Explainer
except ImportError:
    raise ImportError("Package 'gabarit_package' not found. Please install it.")
{%- endif %}



# Manage paths
CURRENT_DIR = Path()
DEFAULT_DATA_DIR = CURRENT_DIR / "{{package_name}}-data"
DEFAULT_MODELS_DIR = CURRENT_DIR / "{{package_name}}-models"

logger = logging.getLogger(__name__)


class ModelSettings(BaseSettings):
    """Download settings

    This class is used for settings management purpose, have a look at the pydantic
    documentation for more details : https://pydantic-docs.helpmanual.io/usage/settings/

    By default, it looks for environment variables (case insensitive) to set the settings
    if a variable is not found, it looks for a file name .env in your working directory
    where you can declare the values of the variables and finally it sets the values
    to the default ones you can see above.
    """

    data_dir: Path = DEFAULT_DATA_DIR
    models_dir: Path = DEFAULT_MODELS_DIR
    model_path: Path = DEFAULT_MODELS_DIR / "model"
    artifactory_model_url: str = ""
    artifactory_user: str = ""
    artifactory_password: str = ""
    redownload: bool = False

    class Config:
        env_file = ".env"


class ModelGabarit(Model):
    """Model class for a Gabarit generated project

    - download_model has been redefined to download a model from artifactory based on
    the settings : ARTIFACTORY_MODEL_URL, ARTIFACTORY_USER, ARTIFACTORY_PASSWORD
    - _load_model has been redefined to use utils_models.load_model
    - predict has been redefined to use utils_models.predict
    """
    def __init__(self, *args, **kwargs):
        """Object initialization
        By default, it initialize the attributes _model, _model_config and _loaded

        see the parent __init__ method in {{package_name}}.model.model_base.Model
        """
        super().__init__(*args, **kwargs)

    def predict(self, content: Any, *args, **kwargs) -> Any:
        """Make a prediction by calling utils_models.predict with the loaded model"""
        if isinstance(content, list) or isinstance(content, dict):
            content = pd.DataFrame(content)

        return utils_models.predict(content, model=self._model, model_conf=self._model_conf, **kwargs)

    def explain_as_json(self, content: Any, *args, **kwargs) -> Union[dict, list]:
        """Compute explanations about a prediction and return a JSON serializable object"""
        if isinstance(content, list) or isinstance(content, dict):
            content = pd.DataFrame(content)

        return self._model_explainer.explain_instance_as_json(content, *args, **kwargs)

    def explain_as_html(self, content: Any, *args, **kwargs) -> str:
        """Compute explanations about a prediction and return an HTML report"""
        if isinstance(content, list) or isinstance(content, dict):
            content = pd.DataFrame(content)

        return self._model_explainer.explain_instance_as_html(content, *args, **kwargs)

    def _load_model(self, **kwargs) -> None:
        """Load a model in a gabarit fashion"""
        settings = ModelSettings(**kwargs)

        # Replace get_data_path method from gabarit.utils to use {{package_name}} data directory
        if hasattr(utils_gabarit, "get_data_path"):
            utils_gabarit.get_data_path = lambda: str(settings.data_dir.resolve())

        # Using is_path=True allow to specify a path instead of a folder relative
        # to gabarit_package.utils.DIR_PATH
        model, model_conf = utils_models.load_model(model_dir=settings.model_path, is_path=True)

        # Set attributes
        self._model = model
        self._model_conf = model_conf

        # Create a model explainer
        self._model_explainer = Explainer(model=model, model_conf=model_conf)

    @staticmethod
    def download_model(**kwargs) -> bool:
        """Download the model from an JFrog Artifactory repository"""
        settings = ModelSettings(**kwargs)

        model_path = settings.model_path

        # If the model already exists there is no need to download it
        if not settings.redownload and model_path.is_dir() and not any(model_path.iterdir()):
            logger.info(f"The model is already dowloaded : {model_path} already exists")
            return True

        # Create models directory if it doesn not exists
        models_dir = settings.models_dir
        models_dir.mkdir(parents=True, exist_ok=True)

        # Download model from artifactory
        try:
            from artifactory import ArtifactoryPath
        except ImportError:
            raise ImportError("Module artifactory not found. Please install it : `pip install dohq-artifactory`")

        model_artifactory_path = ArtifactoryPath(
            settings.artifactory_model_url,
            auth=(settings.artifactory_user, settings.artifactory_password),
            verify=False,
        )

        with tempfile.TemporaryDirectory(dir=models_dir) as tmpdir:
            model_archive_path = Path(tmpdir) / model_artifactory_path.name

            # Download model
            logger.info(f"Downloading the model to : {model_path}")
            with model_archive_path.open("wb") as out:
                model_artifactory_path.writeto(out)

            # Unzip model
            shutil.unpack_archive(model_archive_path, model_path)
            logger.info(f"Model downloaded")

        logger.info(f"Model archive removed")
        return True



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette_prometheus import metrics, PrometheusMiddleware

from .core.config import settings
from .routers import main_routeur
from .core.resources import lifespan


def declare_application() -> FastAPI:
    """Create the FastAPI application

    See https://fastapi.tiangolo.com/tutorial/first-steps/ to learn how to
    customize your FastAPI application
    """
    app = FastAPI(
        title=f"REST API form {settings.app_name}",
        description=f"Use {settings.app_name} thanks to FastAPI",
        lifespan=lifespan
    )

    # Add PrometheusMiddleware
    app.add_middleware(PrometheusMiddleware)
    app.add_route("/metrics", metrics)

    # CORS middleware that allows all origins to avoid CORS problems
    # see https://fastapi.tiangolo.com/tutorial/cors/#use-corsmiddleware
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    #
    app.include_router(main_routeur, prefix=settings.api_entrypoint)

    return app


app = declare_application()
